"""Data registration for Konic Cloud Platform.

This module provides the data registration mechanism that allows agents
to declare data dependencies. These dependencies are automatically
downloaded and made available as environment variables at training time.
"""

from __future__ import annotations

from dataclasses import dataclass
from typing import Literal

from konic.common.errors.cli import KonicValidationError

VersionType = str | Literal["latest"]


@dataclass(frozen=True)
class DataDependency:
    """Represents a registered data dependency.

    Attributes:
        cloud_name: The name of the dataset in Konic Cloud Platform
        env_var: The environment variable name to use for the file path
        version: The version string or "latest" for the current version
    """

    cloud_name: str
    env_var: str
    version: str


_REGISTERED_DATA: list[DataDependency] = []


def register_data(
    cloud_name: str,
    env_var: str,
    version: VersionType = "latest",
) -> DataDependency:
    """
    Register a data dependency for the agent.

    This function declares that the agent requires a specific dataset
    from the Konic Cloud Platform. At training time, the runner will:

    1. Download the dataset (or use cached version if available)
    2. Verify the SHA256 checksum
    3. Set the environment variable to the local file path

    Args:
        cloud_name: The name of the dataset in Konic Cloud Platform.
            Must match an existing dataset uploaded via `konic data push`.
        env_var: The environment variable name that will contain the
            path to the downloaded file. Use this in your environment
            code to access the data.
        version: The version of the dataset to use. Use "latest" to
            always get the most recent version, or specify a specific
            version string (e.g., "1.0.0", "v2", "2024-01-15").

    Returns:
        The created DataDependency object

    Raises:
        KonicValidationError: If cloud_name or env_var is empty

    Example:
        >>> from konic.runtime import register_data
        >>>
        >>> # Register specific version
        >>> register_data("stock-prices", "STOCK_DATA_PATH", "1.0.0")
        >>>
        >>> # Register latest version
        >>> register_data("market-indicators", "INDICATORS_PATH", "latest")
        >>>
        >>> # In your environment code:
        >>> import os
        >>> data_path = os.environ.get("STOCK_DATA_PATH")
        >>> df = pd.read_csv(data_path)
    """
    if not cloud_name:
        raise KonicValidationError(
            "Data cloud_name is required.",
            field="cloud_name",
        )

    if not env_var:
        raise KonicValidationError(
            "Data env_var is required.",
            field="env_var",
        )

    normalized_env_var = env_var.upper().replace("-", "_").replace(" ", "_")

    dependency = DataDependency(
        cloud_name=cloud_name,
        env_var=normalized_env_var,
        version=version,
    )

    _REGISTERED_DATA.append(dependency)
    return dependency


def get_registered_data() -> list[DataDependency]:
    """
    Get all registered data dependencies.

    This function is used internally by the Konic runner to retrieve
    the list of data dependencies that need to be downloaded before
    training starts.

    Returns:
        A list of DataDependency objects. Returns an empty list if
        no data dependencies have been registered.

    Example:
        >>> from konic.runtime import register_data, get_registered_data
        >>>
        >>> register_data("prices", "PRICES_PATH", "1.0.0")
        >>> register_data("indicators", "INDICATORS_PATH", "latest")
        >>>
        >>> deps = get_registered_data()
        >>> for dep in deps:
        ...     print(f"{dep.cloud_name} -> ${dep.env_var} (v{dep.version})")
        prices -> $PRICES_PATH (v1.0.0)
        indicators -> $INDICATORS_PATH (vlatest)
    """
    return _REGISTERED_DATA.copy()


def clear_registered_data() -> None:
    """
    Clear all registered data dependencies.

    This function is primarily used for testing purposes to reset
    the global state between tests.
    """
    global _REGISTERED_DATA
    _REGISTERED_DATA = []



import inspect
from collections.abc import Callable

from konic.environment.reward.enums import KonicRewardKeys


def get_custom_reward_fns(obj) -> list[Callable[..., float]]:
    """Retrieve a list of custom reward functions from the given object using introspection."""
    members = inspect.getmembers(obj, predicate=inspect.ismethod)
    return [fn for _, fn in members if hasattr(fn, KonicRewardKeys.CUSTOM_REWARD_FN_ATTR_KEY.value)]



import os
from pathlib import Path

from cookiecutter.main import cookiecutter
from typer import Typer

from konic.cli.decorators import error_handler, pretty_output
from konic.cli.enums import BoilerplateOptions
from konic.cli.env_keys import KonicCLIEnvVars
from konic.cli.src import agents, artifacts, data, health, training
from konic.cli.utils import compile_artifact

app = Typer(
    name="konic",
    help="Official Konic Agent Development Toolkit",
    no_args_is_help=True,
)

app.add_typer(health.app, name="health")
app.add_typer(agents.app, name="agent")
app.add_typer(training.app, name="train")
app.add_typer(data.app, name="data")
app.add_typer(artifacts.app, name="artifact")


@app.command(name="get-host")
@error_handler(exit_on_error=True)
@pretty_output(title="Konic Host Configuration", border_style="cyan")
def get_host() -> str:
    """
    Get the Konic host URL from environment variables
    """
    host = os.environ.get(KonicCLIEnvVars.KONIC_HOST.value, "")
    return host if host else "Not configured"


@app.command(name="init")
@error_handler(exit_on_error=True)
def create_boilerplate(template: BoilerplateOptions = BoilerplateOptions.basic):
    path = Path(__file__).parent / "boilerplates" / template.value
    cookiecutter(str(path))


@app.command(name="compile")
@error_handler(exit_on_error=True)
def compile(filepath: str):
    compile_artifact(filepath)



from konic.environment.termination.decorators import custom_termination
from konic.environment.termination.termination import KonicTerminationComposer

__all__ = ["KonicTerminationComposer", "custom_termination"]



from abc import ABC, abstractmethod

import gymnasium as gym
from pydantic.fields import FieldInfo
from pydantic_core import PydanticUndefinedType


class BaseKonicSpaceTypeStrategy(ABC):
    @classmethod
    @abstractmethod
    def apply(cls, name: str, field: FieldInfo) -> dict[str, gym.spaces.Space]:
        pass

    @staticmethod
    def _has_default(field: FieldInfo) -> bool:
        """Check if the field has a default value set."""
        return not isinstance(field.default, PydanticUndefinedType)



"""CLI commands for managing artifacts on Konic Cloud Platform."""

from __future__ import annotations

import json
from pathlib import Path
from typing import Annotated

import typer
from rich.console import Console
from rich.panel import Panel
from rich.table import Table

from konic.cli.client import client
from konic.cli.decorators import error_handler, loading_indicator
from konic.cli.env_keys import KonicCLIEnvVars
from konic.cli.utils import resolve_agent_identifier
from konic.common.errors import (
    KonicArtifactNotFoundError,
    KonicHTTPError,
)

app = typer.Typer(
    name="artifact",
    help="Manage training artifacts (checkpoints and models) on Konic Cloud Platform",
    no_args_is_help=True,
)

console = Console()


def _apply_host_override(host: str | None) -> None:
    """Apply host override if provided."""
    if host:
        client.set_base_url(host)


def _format_file_size(size_bytes: int) -> str:
    """Format file size in human-readable format."""
    size = float(size_bytes)
    for unit in ["B", "KB", "MB", "GB"]:
        if size < 1024:
            return f"{size:.1f} {unit}"
        size /= 1024
    return f"{size:.1f} TB"


def _get_artifact_type_style(artifact_type: str) -> str:
    """Get Rich style for artifact type."""
    styles = {
        "final": "bold green",
        "checkpoint": "cyan",
    }
    return styles.get(artifact_type, "white")


def _display_artifact_table(artifacts: list[dict], agent_name: str | None = None) -> None:
    """Display artifacts in a Rich table."""
    title = "Artifacts"
    if agent_name:
        title = f"Artifacts for agent '{agent_name}'"

    table = Table(
        title=title,
        show_header=True,
        header_style="bold magenta",
    )
    table.add_column("ID", style="dim")
    table.add_column("Type")
    table.add_column("Iteration", justify="right")
    table.add_column("Job ID", style="dim")
    table.add_column("Size", justify="right")
    table.add_column("Created At")

    for artifact in artifacts:
        artifact_type = artifact.get("artifact_type", "unknown")
        job_id = artifact.get("training_job_id", "N/A")
        job_id_display = job_id

        table.add_row(
            artifact.get("id", "N/A"),
            f"[{_get_artifact_type_style(artifact_type)}]{artifact_type}[/]",
            str(artifact.get("iteration", "N/A")),
            job_id_display,
            _format_file_size(artifact.get("file_size", 0)),
            artifact.get("created_at", "N/A")[:19].replace("T", " "),
        )

    console.print(table)


def _display_artifact_detail(artifact: dict) -> None:
    """Display artifact details in a Rich panel."""
    artifact_type = artifact.get("artifact_type", "unknown")

    content = f"""[bold cyan]ID:[/bold cyan] {artifact.get('id', 'N/A')}
[bold cyan]Agent:[/bold cyan] {artifact.get('agent_name', 'N/A')} ({artifact.get('agent_id', 'N/A')})
[bold cyan]Job ID:[/bold cyan] {artifact.get('training_job_id', 'N/A')}
[bold cyan]Type:[/bold cyan] [{_get_artifact_type_style(artifact_type)}]{artifact_type}[/]
[bold cyan]Iteration:[/bold cyan] {artifact.get('iteration', 'N/A')}
[bold cyan]Size:[/bold cyan] {_format_file_size(artifact.get('file_size', 0))}
[bold cyan]Checksum (SHA256):[/bold cyan] {artifact.get('checksum_sha256', 'N/A')}
[bold cyan]Created:[/bold cyan] {artifact.get('created_at', 'N/A')[:19].replace('T', ' ')}"""

    console.print(Panel(content, title="Artifact Details", border_style="blue"))


@app.command(name="list")
@error_handler(exit_on_error=True)
@loading_indicator(message="Fetching artifacts...")
def list_artifacts(
    agent: Annotated[
        str,
        typer.Argument(
            help="Agent ID or name",
        ),
    ],
    job_id: Annotated[
        str | None,
        typer.Option(
            "--job-id",
            "-j",
            help="Filter by training job ID",
        ),
    ] = None,
    artifact_type: Annotated[
        str | None,
        typer.Option(
            "--type",
            "-t",
            help="Filter by artifact type (checkpoint or final)",
        ),
    ] = None,
    host: Annotated[
        str | None,
        typer.Option(
            "--host",
            "-h",
            help="Override the Konic API host URL",
            envvar=KonicCLIEnvVars.KONIC_HOST.value,
        ),
    ] = None,
    output_json: Annotated[
        bool,
        typer.Option(
            "--json",
            help="Output as JSON",
        ),
    ] = False,
) -> None:
    """
    List artifacts for an agent.

    Example:
        konic artifact list my-agent
        konic artifact list my-agent --type final
        konic artifact list my-agent --job-id 692bba9e30c77da5144f5a95
    """
    _apply_host_override(host)

    agent_id = resolve_agent_identifier(client, agent)

    try:
        agent_info = client.get_json(f"/agents/{agent_id}")
        agent_name = agent_info.get("agent_name", agent)
    except KonicHTTPError:
        agent_name = agent

    if job_id:
        result = client.get_json(f"/artifacts/by-job/{job_id}")
    else:
        result = client.get_json(f"/artifacts/by-agent/{agent_id}")

    if artifact_type:
        result = [a for a in result if a.get("artifact_type") == artifact_type]

    if output_json:
        console.print_json(json.dumps(result))
        return

    if not result:
        console.print("[yellow]No artifacts found.[/yellow]")
        return

    _display_artifact_table(result, agent_name)


@app.command(name="show")
@error_handler(exit_on_error=True)
@loading_indicator(message="Fetching artifact details...")
def show_artifact(
    artifact_id: Annotated[
        str,
        typer.Argument(
            help="Artifact ID",
        ),
    ],
    host: Annotated[
        str | None,
        typer.Option(
            "--host",
            "-h",
            help="Override the Konic API host URL",
            envvar=KonicCLIEnvVars.KONIC_HOST.value,
        ),
    ] = None,
    output_json: Annotated[
        bool,
        typer.Option(
            "--json",
            help="Output as JSON",
        ),
    ] = False,
) -> None:
    """
    Show details of an artifact.

    Example:
        konic artifact show artifact-abc12345-1701388800
    """
    _apply_host_override(host)

    try:
        result = client.get_json(f"/artifacts/{artifact_id}")

        if output_json:
            console.print_json(json.dumps(result))
            return

        _display_artifact_detail(result)
    except KonicHTTPError as e:
        if e.status_code == 404:
            raise KonicArtifactNotFoundError(artifact_id)
        raise


@app.command(name="download")
@error_handler(exit_on_error=True)
def download_artifact(
    artifact_id: Annotated[
        str,
        typer.Argument(
            help="Artifact ID to download",
        ),
    ],
    output: Annotated[
        Path | None,
        typer.Option(
            "--output",
            "-o",
            help="Output directory (default: current directory)",
            file_okay=False,
            dir_okay=True,
            resolve_path=True,
        ),
    ] = None,
    skip_verify: Annotated[
        bool,
        typer.Option(
            "--skip-verify",
            help="Skip SHA256 checksum verification",
        ),
    ] = False,
    host: Annotated[
        str | None,
        typer.Option(
            "--host",
            "-h",
            help="Override the Konic API host URL",
            envvar=KonicCLIEnvVars.KONIC_HOST.value,
        ),
    ] = None,
) -> None:
    """
    Download an artifact.

    Downloads the artifact and verifies the SHA256 checksum by default.

    Example:
        konic artifact download artifact-abc12345-1701388800
        konic artifact download artifact-abc12345-1701388800 --output ./checkpoints/
        konic artifact download artifact-abc12345-1701388800 --skip-verify
    """
    _apply_host_override(host)

    try:
        download_info = client.get_json(f"/artifacts/{artifact_id}/download")
    except KonicHTTPError as e:
        if e.status_code == 404:
            raise KonicArtifactNotFoundError(artifact_id)
        raise

    output_dir = output or Path.cwd()
    output_dir.mkdir(parents=True, exist_ok=True)

    artifact_type = download_info.get("artifact_type", "artifact")
    iteration = download_info.get("iteration", 0)
    filename = f"{artifact_type}_{iteration:06d}.zip"
    output_path = output_dir / filename

    console.print(f"[cyan]Downloading {artifact_type} artifact (iteration {iteration})...[/cyan]\n")

    try:
        downloaded_path, checksum = client.download_file_with_progress(
            url=download_info["download_url"],
            output_path=output_path,
            expected_size=download_info.get("file_size"),
            expected_checksum=download_info.get("checksum_sha256"),
            verify_checksum=not skip_verify,
        )

        console.print("\n[bold green]✓[/bold green] Downloaded successfully!")
        console.print(f"  [cyan]File:[/cyan] {downloaded_path}")
        console.print(
            f"  [cyan]Size:[/cyan] {_format_file_size(download_info.get('file_size', 0))}"
        )
        console.print(f"  [cyan]Checksum:[/cyan] {checksum}")

        if skip_verify:
            console.print("  [yellow]⚠ Checksum verification was skipped[/yellow]")
        else:
            console.print("  [green]✓ Checksum verified[/green]")

    except ValueError as e:
        console.print(f"\n[bold red]✗[/bold red] {e}")
        raise typer.Exit(1)



import functools
import hashlib
import json
import os
from collections.abc import Callable, Generator
from pathlib import Path
from typing import Any, TypeVar

import httpx
from dotenv import load_dotenv
from rich.progress import (
    BarColumn,
    DownloadColumn,
    Progress,
    TaskProgressColumn,
    TextColumn,
    TimeRemainingColumn,
    TransferSpeedColumn,
)

from konic.cli.env_keys import KonicCLIEnvVars
from konic.common.errors import KonicEnvironmentError, KonicHTTPError

F = TypeVar("F", bound=Callable[..., Any])


def handle_http_errors[F: Callable[..., Any]](func: F) -> F:
    """
    Decorator to handle HTTP errors for API client methods.
    Converts httpx.HTTPStatusError into KonicHTTPError with detailed information.
    """

    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        response = func(*args, **kwargs)
        try:
            response.raise_for_status()
            return response
        except httpx.HTTPStatusError as e:
            try:
                error_detail = response.json().get("detail", str(e))
            except Exception:
                error_detail = response.text or str(e)

            raise KonicHTTPError(
                message=error_detail,
                status_code=response.status_code,
                endpoint=str(response.url),
                response_body=response.text[:500] if response.text else None,
            )

    return wrapper  # type: ignore


class KonicAPIClient:
    def __init__(self):
        load_dotenv()

        headers = {"Accept": "application/json", "User-Agent": "konic-cli/0.1.0"}

        self.BASE_URL = os.environ.get(KonicCLIEnvVars.KONIC_HOST.value, "")
        if self.BASE_URL is None or self.BASE_URL == "":
            raise KonicEnvironmentError(
                env_var=KonicCLIEnvVars.KONIC_HOST.value,
                suggestion=f"export {KonicCLIEnvVars.KONIC_HOST.value}=<host-url>",
            )

        self.APIKEY = None  # TODO: not-implemented yet
        self.client = httpx.Client(base_url=self.BASE_URL, headers=headers)

    @handle_http_errors
    def get(
        self,
        endpoint: str,
        params: dict[str, Any] | None = None,
        headers: dict[str, str] | None = None,
    ) -> httpx.Response:
        """
        Perform a GET request.

        Args:
            endpoint: The API endpoint to call
            params: Optional query parameters
            headers: Optional additional headers

        Returns:
            httpx.Response: The response object

        Raises:
            KonicHTTPError: If the request fails
        """
        return self.client.get(endpoint, params=params, headers=headers)

    @handle_http_errors
    def post(
        self,
        endpoint: str,
        json: dict[str, Any] | None = None,
        data: dict[str, Any] | None = None,
        headers: dict[str, str] | None = None,
    ) -> httpx.Response:
        """
        Perform a POST request.

        Args:
            endpoint: The API endpoint to call
            json: Optional JSON body
            data: Optional form data
            headers: Optional additional headers

        Returns:
            httpx.Response: The response object

        Raises:
            KonicHTTPError: If the request fails
        """
        return self.client.post(endpoint, json=json, data=data, headers=headers)

    @handle_http_errors
    def put(
        self,
        endpoint: str,
        json: dict[str, Any] | None = None,
        data: dict[str, Any] | None = None,
        headers: dict[str, str] | None = None,
    ) -> httpx.Response:
        """
        Perform a PUT request.

        Args:
            endpoint: The API endpoint to call
            json: Optional JSON body
            data: Optional form data
            headers: Optional additional headers

        Returns:
            httpx.Response: The response object

        Raises:
            KonicHTTPError: If the request fails
        """
        return self.client.put(endpoint, json=json, data=data, headers=headers)

    @handle_http_errors
    def patch(
        self,
        endpoint: str,
        json: dict[str, Any] | None = None,
        data: dict[str, Any] | None = None,
        headers: dict[str, str] | None = None,
    ) -> httpx.Response:
        """
        Perform a PATCH request.

        Args:
            endpoint: The API endpoint to call
            json: Optional JSON body
            data: Optional form data
            headers: Optional additional headers

        Returns:
            httpx.Response: The response object

        Raises:
            KonicHTTPError: If the request fails
        """
        return self.client.patch(endpoint, json=json, data=data, headers=headers)

    @handle_http_errors
    def delete(
        self,
        endpoint: str,
        params: dict[str, Any] | None = None,
        headers: dict[str, str] | None = None,
    ) -> httpx.Response:
        """
        Perform a DELETE request.

        Args:
            endpoint: The API endpoint to call
            params: Optional query parameters
            headers: Optional additional headers

        Returns:
            httpx.Response: The response object

        Raises:
            KonicHTTPError: If the request fails
        """
        return self.client.delete(endpoint, params=params, headers=headers)

    def get_json(
        self,
        endpoint: str,
        params: dict[str, Any] | None = None,
        headers: dict[str, str] | None = None,
    ) -> Any:
        """
        Perform a GET request and return JSON response.

        Args:
            endpoint: The API endpoint to call
            params: Optional query parameters
            headers: Optional additional headers

        Returns:
            Any: The parsed JSON response

        Raises:
            KonicHTTPError: If the request fails
        """
        response = self.get(endpoint, params=params, headers=headers)
        return response.json()

    def post_json(
        self,
        endpoint: str,
        json: dict[str, Any] | None = None,
        data: dict[str, Any] | None = None,
        headers: dict[str, str] | None = None,
    ) -> Any:
        """
        Perform a POST request and return JSON response.

        Args:
            endpoint: The API endpoint to call
            json: Optional JSON body
            data: Optional form data
            headers: Optional additional headers

        Returns:
            Any: The parsed JSON response

        Raises:
            KonicHTTPError: If the request fails
        """
        response = self.post(endpoint, json=json, data=data, headers=headers)
        return response.json()

    def put_json(
        self,
        endpoint: str,
        json: dict[str, Any] | None = None,
        data: dict[str, Any] | None = None,
        headers: dict[str, str] | None = None,
    ) -> Any:
        """
        Perform a PUT request and return JSON response.

        Args:
            endpoint: The API endpoint to call
            json: Optional JSON body
            data: Optional form data
            headers: Optional additional headers

        Returns:
            Any: The parsed JSON response

        Raises:
            KonicHTTPError: If the request fails
        """
        response = self.put(endpoint, json=json, data=data, headers=headers)
        return response.json()

    def patch_json(
        self,
        endpoint: str,
        json: dict[str, Any] | None = None,
        data: dict[str, Any] | None = None,
        headers: dict[str, str] | None = None,
    ) -> Any:
        """
        Perform a PATCH request and return JSON response.

        Args:
            endpoint: The API endpoint to call
            json: Optional JSON body
            data: Optional form data
            headers: Optional additional headers

        Returns:
            Any: The parsed JSON response

        Raises:
            KonicHTTPError: If the request fails
        """
        response = self.patch(endpoint, json=json, data=data, headers=headers)
        return response.json()

    def delete_json(
        self,
        endpoint: str,
        params: dict[str, Any] | None = None,
        headers: dict[str, str] | None = None,
    ) -> Any:
        """
        Perform a DELETE request and return JSON response.

        Args:
            endpoint: The API endpoint to call
            params: Optional query parameters
            headers: Optional additional headers

        Returns:
            Any: The parsed JSON response

        Raises:
            KonicHTTPError: If the request fails
        """
        response = self.delete(endpoint, params=params, headers=headers)
        return response.json()

    def stream_sse(
        self,
        endpoint: str,
        params: dict[str, Any] | None = None,
        headers: dict[str, str] | None = None,
        timeout: float = 300.0,
    ) -> Generator[dict[str, Any], None, None]:
        """
        Stream Server-Sent Events from an endpoint.

        Args:
            endpoint: The API endpoint to stream from
            params: Optional query parameters
            headers: Optional additional headers
            timeout: Timeout in seconds (default: 300)

        Yields:
            dict: Parsed event data with 'event' and 'data' keys

        Raises:
            KonicHTTPError: If the request fails
        """
        stream_headers = {"Accept": "text/event-stream"}
        if headers:
            stream_headers.update(headers)

        with httpx.Client(
            base_url=self.BASE_URL,
            timeout=httpx.Timeout(timeout, connect=10.0),
        ) as stream_client:
            with stream_client.stream(
                "GET",
                endpoint,
                params=params,
                headers=stream_headers,
            ) as response:
                try:
                    response.raise_for_status()
                except httpx.HTTPStatusError:
                    try:
                        error_detail = response.read().decode()
                    except Exception:
                        error_detail = str(response.status_code)

                    raise KonicHTTPError(
                        message=error_detail,
                        status_code=response.status_code,
                        endpoint=str(response.url),
                    )

                event_type = None
                for line in response.iter_lines():
                    line = line.strip()
                    if not line:
                        continue

                    if line.startswith("event:"):
                        event_type = line[6:].strip()
                    elif line.startswith("data:"):
                        data_str = line[5:].strip()
                        try:
                            data = json.loads(data_str) if data_str else {}
                        except json.JSONDecodeError:
                            data = {"raw": data_str}

                        yield {"event": event_type or "message", "data": data}
                        event_type = None

    def close(self):
        """Close the HTTP client connection."""
        self.client.close()

    def __enter__(self):
        """Context manager entry."""
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        """Context manager exit."""
        self.close()

    def set_base_url(self, base_url: str) -> None:
        """
        Update the base URL for the client.

        Args:
            base_url: The new base URL to use for requests
        """
        self.BASE_URL = base_url
        headers = {"Accept": "application/json", "User-Agent": "konic-cli/0.1.0"}
        self.client = httpx.Client(base_url=self.BASE_URL, headers=headers)

    @handle_http_errors
    def upload_file(
        self,
        endpoint: str,
        file_path: Path,
        headers: dict[str, str] | None = None,
    ) -> httpx.Response:
        """
        Upload a file via multipart/form-data POST request.

        Args:
            endpoint: The API endpoint to call
            file_path: Path to the file to upload
            headers: Optional additional headers

        Returns:
            httpx.Response: The response object

        Raises:
            KonicHTTPError: If the request fails
            FileNotFoundError: If the file does not exist
        """
        with open(file_path, "rb") as f:
            files = {"file": (file_path.name, f, "application/octet-stream")}
            return self.client.post(endpoint, files=files, headers=headers)

    def upload_file_json(
        self,
        endpoint: str,
        file_path: Path,
        headers: dict[str, str] | None = None,
    ) -> Any:
        """
        Upload a file and return JSON response.

        Args:
            endpoint: The API endpoint to call
            file_path: Path to the file to upload
            headers: Optional additional headers

        Returns:
            Any: The parsed JSON response

        Raises:
            KonicHTTPError: If the request fails
        """
        response = self.upload_file(endpoint, file_path, headers=headers)
        return response.json()

    def download_file(
        self,
        endpoint: str,
        output_path: Path,
        params: dict[str, Any] | None = None,
        headers: dict[str, str] | None = None,
    ) -> Path:
        """
        Download a file from the API and save it to disk.

        Args:
            endpoint: The API endpoint to call
            output_path: Path where the file should be saved
            params: Optional query parameters
            headers: Optional additional headers

        Returns:
            Path: The path to the downloaded file

        Raises:
            KonicHTTPError: If the request fails
        """
        with self.client.stream("GET", endpoint, params=params, headers=headers) as response:
            try:
                response.raise_for_status()
            except httpx.HTTPStatusError:
                try:
                    error_detail = response.json().get("detail", str(response.status_code))
                except Exception:
                    error_detail = response.text or str(response.status_code)

                raise KonicHTTPError(
                    message=error_detail,
                    status_code=response.status_code,
                    endpoint=str(response.url),
                    response_body=response.text[:500] if response.text else None,
                )

            output_path.parent.mkdir(parents=True, exist_ok=True)
            with open(output_path, "wb") as f:
                for chunk in response.iter_bytes():
                    f.write(chunk)

        return output_path

    @handle_http_errors
    def upload_file_with_form_data(
        self,
        endpoint: str,
        file_path: Path,
        form_data: dict[str, str],
        headers: dict[str, str] | None = None,
    ) -> httpx.Response:
        """
        Upload a file with additional form data via multipart/form-data POST.

        Args:
            endpoint: The API endpoint to call
            file_path: Path to the file to upload
            form_data: Additional form fields to include (e.g., name, version)
            headers: Optional additional headers

        Returns:
            httpx.Response: The response object

        Raises:
            KonicHTTPError: If the request fails
            FileNotFoundError: If the file does not exist
        """
        with open(file_path, "rb") as f:
            files = {"file": (file_path.name, f, "application/octet-stream")}
            return self.client.post(
                endpoint,
                files=files,
                data=form_data,
                headers=headers,
            )

    def upload_file_with_form_data_json(
        self,
        endpoint: str,
        file_path: Path,
        form_data: dict[str, str],
        headers: dict[str, str] | None = None,
    ) -> Any:
        """
        Upload a file with form data and return JSON response.

        Args:
            endpoint: The API endpoint to call
            file_path: Path to the file to upload
            form_data: Additional form fields to include
            headers: Optional additional headers

        Returns:
            Any: The parsed JSON response

        Raises:
            KonicHTTPError: If the request fails
        """
        response = self.upload_file_with_form_data(endpoint, file_path, form_data, headers=headers)
        return response.json()

    def upload_file_with_progress(
        self,
        endpoint: str,
        file_path: Path,
        form_data: dict[str, str],
        headers: dict[str, str] | None = None,
    ) -> Any:
        """
        Upload a file with progress bar and return JSON response.

        Args:
            endpoint: The API endpoint to call
            file_path: Path to the file to upload
            form_data: Additional form fields to include
            headers: Optional additional headers

        Returns:
            Any: The parsed JSON response

        Raises:
            KonicHTTPError: If the request fails
        """
        file_size = file_path.stat().st_size

        with Progress(
            TextColumn("[bold blue]{task.description}"),
            BarColumn(),
            TaskProgressColumn(),
            TransferSpeedColumn(),
            TimeRemainingColumn(),
        ) as progress:
            task = progress.add_task(f"Uploading {file_path.name}", total=file_size)

            with open(file_path, "rb") as f:
                content = f.read()
                progress.update(task, completed=file_size)

            files = {"file": (file_path.name, content, "application/octet-stream")}
            response = self.client.post(
                endpoint,
                files=files,
                data=form_data,
                headers=headers,
            )

            try:
                response.raise_for_status()
            except httpx.HTTPStatusError:
                try:
                    error_detail = response.json().get("detail", str(response))
                except Exception:
                    error_detail = response.text or str(response.status_code)

                raise KonicHTTPError(
                    message=error_detail,
                    status_code=response.status_code,
                    endpoint=str(response.url),
                    response_body=response.text[:500] if response.text else None,
                )

        return response.json()

    def download_file_with_progress(
        self,
        url: str,
        output_path: Path,
        expected_size: int | None = None,
        expected_checksum: str | None = None,
        verify_checksum: bool = True,
    ) -> tuple[Path, str]:
        """
        Download a file from a URL with progress bar and optional checksum verification.

        Args:
            url: The full URL to download from (e.g., presigned S3 URL)
            output_path: Path where the file should be saved
            expected_size: Expected file size in bytes for progress bar
            expected_checksum: Expected SHA256 checksum for verification
            verify_checksum: Whether to verify the checksum after download

        Returns:
            tuple[Path, str]: The path to the downloaded file and its SHA256 checksum

        Raises:
            KonicHTTPError: If the request fails
            ValueError: If checksum verification fails
        """
        output_path.parent.mkdir(parents=True, exist_ok=True)
        sha256_hash = hashlib.sha256()

        with Progress(
            TextColumn("[bold blue]{task.description}"),
            BarColumn(),
            DownloadColumn(),
            TransferSpeedColumn(),
            TimeRemainingColumn(),
        ) as progress:
            with httpx.Client(timeout=httpx.Timeout(300.0, connect=10.0)) as download_client:
                with download_client.stream("GET", url) as response:
                    try:
                        response.raise_for_status()
                    except httpx.HTTPStatusError:
                        raise KonicHTTPError(
                            message=f"Failed to download file: {response.status_code}",
                            status_code=response.status_code,
                            endpoint=url,
                        )

                    total_size = expected_size
                    if total_size is None:
                        content_length = response.headers.get("content-length")
                        if content_length:
                            total_size = int(content_length)

                    task = progress.add_task(
                        f"Downloading {output_path.name}",
                        total=total_size,
                    )

                    with open(output_path, "wb") as f:
                        for chunk in response.iter_bytes(chunk_size=8192):
                            f.write(chunk)
                            sha256_hash.update(chunk)
                            progress.update(task, advance=len(chunk))

        actual_checksum = sha256_hash.hexdigest()

        if verify_checksum and expected_checksum:
            if actual_checksum != expected_checksum:
                output_path.unlink(missing_ok=True)
                raise ValueError(
                    f"Checksum verification failed!\n"
                    f"Expected: {expected_checksum}\n"
                    f"Actual:   {actual_checksum}"
                )

        return output_path, actual_checksum



"""
Counter Environment Example

A simple environment where an agent tries to reach a target count.
The agent can increment or decrement a counter, starting from 0.

Actions:
    0: Decrement (-1)
    1: Increment (+1)

Goal:
    Reach the target count (default: 10)

Reward:
    - Positive reward for getting closer to the target
    - Negative reward for moving away from the target
    - Large bonus reward when reaching the target

Termination:
    - Episode ends when the target is reached
    - Episode ends after max_steps (default: 50)
"""

from __future__ import annotations

from typing import Any

import numpy as np

from konic.environment import KonicEnvironment
from konic.environment.reward import KonicRewardComposer, custom_reward
from konic.environment.space import KonicBound, KonicDiscrete, KonicSpace
from konic.environment.termination import KonicTerminationComposer, custom_termination


class CounterActionSpace(KonicSpace):
    action: KonicDiscrete = 2


class CounterObservationSpace(KonicSpace):
    current_count: KonicBound = ((1,), -100.0, 100.0)
    target_count: KonicBound = ((1,), -100.0, 100.0)
    steps_remaining: KonicBound = ((1,), 0.0, 100.0)


class CounterRewardComposer(KonicRewardComposer["CounterEnvironment"]):
    @custom_reward
    def target_reached_bonus(self) -> float:
        if self.env.current_count == self.env.target_count:
            return 10.0
        return 0.0

    @custom_reward
    def distance_penalty(self) -> float:
        distance = abs(self.env.current_count - self.env.target_count)
        return -0.01 * distance

    @custom_reward
    def progress_reward(self) -> float:
        if not hasattr(self.env, "previous_distance"):
            return 0.0

        current_distance = abs(self.env.current_count - self.env.target_count)
        previous_distance = self.env.previous_distance

        if current_distance < previous_distance:
            return 1.0
        elif current_distance > previous_distance:
            return -1.0
        return 0.0


class CounterTerminationComposer(KonicTerminationComposer["CounterEnvironment"]):
    def terminated(self) -> bool:
        return False

    @custom_termination
    def target_reached(self) -> bool:
        return self.env.current_count == self.env.target_count

    @custom_termination
    def max_steps_reached(self) -> bool:
        return self.env.steps >= self.env.max_steps


class CounterEnvironment(KonicEnvironment):
    def __init__(self, target_count: int = 10, max_steps: int = 50, flatten_spaces: bool = False):
        """
        Initialize the Counter Environment.

        Args:
            target_count: The target count to reach (default: 10)
            max_steps: Maximum number of steps per episode (default: 50)
            flatten_spaces: If True, automatically flatten Dict spaces for RLlib compatibility (default: False)
        """
        super().__init__(
            action_space=CounterActionSpace(),
            observation_space=CounterObservationSpace(),
            reward_composer=CounterRewardComposer(),
            termination_composer=CounterTerminationComposer(),
            flatten_spaces=flatten_spaces,
        )

        self.steps = 0
        self.current_count = 0
        self.max_steps = max_steps
        self.target_count = target_count
        self.previous_distance = abs(self.current_count - self.target_count)

    def _reset(self, seed: int | None = None, options: dict | None = None):
        if seed is not None:
            np.random.seed(seed)

        self.current_count = 0
        self.steps = 0
        self.previous_distance = abs(self.current_count - self.target_count)

        observation = self.get_obs()
        info = self.get_info()

        return observation, info

    def _step(self, action: dict[str, Any]):
        self.previous_distance = abs(self.current_count - self.target_count)

        action_value = action["action"]
        if action_value == 0:
            self.current_count -= 1
        elif action_value == 1:
            self.current_count += 1

        self.steps += 1

        observation = self.get_obs()
        reward = self.reward_composer.compose()
        terminated = self.termination_composer.compose()
        truncated = False
        info = self.get_info()

        return observation, reward, terminated, truncated, info

    def get_obs(self):
        return {
            "current_count": np.array([self.current_count], dtype=np.float32),
            "target_count": np.array([self.target_count], dtype=np.float32),
            "steps_remaining": np.array([self.max_steps - self.steps], dtype=np.float32),
        }

    def get_info(self):
        return {
            "current_count": self.current_count,
            "target_count": self.target_count,
            "steps": self.steps,
            "distance_to_target": abs(self.current_count - self.target_count),
        }


def main() -> None:
    seed = 42

    print("=== Example 1: Dict observation space ===")
    env_dict = CounterEnvironment(flatten_spaces=False)
    obs, info = env_dict.reset(seed=seed)
    print(f"Observation space: {env_dict.observation_space}")
    print(f"Action space: {env_dict.action_space}")
    print(f"Sample observation: {obs}")
    print()

    print("=== Example 2: Flattened observation space (RLlib-friendly) ===")
    env_flat = CounterEnvironment(flatten_spaces=True)
    obs, info = env_flat.reset(seed=seed)
    print(f"Observation space: {env_flat.observation_space}")
    print(f"Action space: {env_flat.action_space}")
    print(f"Sample observation (flattened): {obs}")
    print()

    print("=== Running 10 steps with flattened environment ===")
    for i in range(10):
        action = env_flat.action_space.sample()
        obs, reward, terminated, truncated, info = env_flat.step(action)
        print(f"Step {i+1}: Action={action}, Obs shape={obs.shape}, Reward={reward:.2f}")

        if terminated or truncated:
            break


if __name__ == "__main__":
    main()



from typing import Any

import gymnasium as gym

from konic.agent.base import BaseKonicAgent
from konic.common.errors import KonicAssertionError, KonicConfigurationError
from konic.environment import KonicEnvironment
from konic.module import BaseTorchModule, KonicTorchPPO


class KonicAgent(BaseKonicAgent):
    def __init__(
        self,
        environment: KonicEnvironment | gym.Env | None = None,
        environment_config: dict[str, Any] | None = None,
        algorithm_config: dict[str, Any] | None = None,
        module: type[BaseTorchModule] | None = KonicTorchPPO,
        training_config: dict[str, Any] | None = None,
    ):
        self._environment = environment
        self._environment_config = environment_config
        self._algorithm_config = algorithm_config
        self._module = module
        self._training_config = training_config

        self._has_environment = environment is not None
        self._has_environment_config = environment_config is not None
        self._has_algorithm_config = algorithm_config is not None
        self._has_training_config = training_config is not None

    def get_environment(self) -> KonicEnvironment | gym.Env:
        """Return the environment class to be used for training."""
        if self._has_environment:
            if self._environment is None:
                raise KonicAssertionError(
                    "Environment should not be None when _has_environment is True"
                )
            return self._environment
        raise KonicConfigurationError(
            "Environment must be provided either in __init__ or by overriding get_environment()",
            config_key="environment",
        )

    def get_environment_config(self) -> dict[str, Any]:
        """Return configuration parameters for environment initialization."""
        if self._has_environment_config:
            if self._environment_config is None:
                raise KonicAssertionError(
                    "Environment config should not be None when _has_environment_config is True"
                )
            return self._environment_config
        return {}

    def get_algorithm_config(self) -> dict[str, Any]:
        """Return algorithm-specific configuration."""
        if self._has_algorithm_config:
            if self._algorithm_config is None:
                raise KonicAssertionError(
                    "Algorithm config should not be None when _has_algorithm_config is True"
                )
            return self._algorithm_config
        return {}

    def get_module(self) -> type[BaseTorchModule]:
        """Return neural network module (defaults to KonicTorchPPO)."""
        if self._module is None:
            raise KonicAssertionError("Module should not be None")
        return self._module

    def get_training_config(self) -> dict[str, Any]:
        """Return training configuration (num_iterations, batch_size, etc.)."""
        if self._has_training_config:
            if self._training_config is None:
                raise KonicAssertionError(
                    "Training config should not be None when _has_training_config is True"
                )
            return self._training_config
        return {}



from typing import Any

import gymnasium as gym
import numpy as np

from konic.environment.base import BaseKonicEnvironment
from konic.environment.reward import KonicRewardComposer
from konic.environment.space import KonicSpace
from konic.environment.termination import KonicTerminationComposer
from konic.environment.type import KonicResetStateObservation, KonicStepStateObservation


class KonicEnvironment(BaseKonicEnvironment):
    """
    Base Konic Environment compatible with RLlib.

    This environment automatically handles Dict observation/action spaces for RLlib compatibility.
    Set `flatten_spaces=True` to automatically flatten Dict spaces to Box spaces, which is
    recommended for most RLlib algorithms.
    """

    def __init__(
        self,
        action_space: KonicSpace,
        observation_space: KonicSpace,
        reward_composer: KonicRewardComposer,
        termination_composer: KonicTerminationComposer,
        flatten_spaces: bool = False,
    ):
        super().__init__()

        self._konic_action_space = action_space
        self._konic_observation_space = observation_space

        self._action_space = action_space.to_gym()
        self._observation_space = observation_space.to_gym()

        self._flatten_spaces = flatten_spaces
        if flatten_spaces:
            self._apply_space_flattening()

        self._reward_composer = reward_composer
        self._reward_composer.set_env(self)

        self._termination_composer = termination_composer
        if self._termination_composer is not None:
            self._termination_composer.set_env(self)

    def _apply_space_flattening(self):
        """Apply FlattenObservation-style flattening to Dict spaces for RLlib compatibility."""
        if isinstance(self._observation_space, gym.spaces.Dict):
            obs_dim = sum(
                int(np.prod(space.shape))
                for space in self._observation_space.spaces.values()
                if isinstance(space, gym.spaces.Box)
            )

            self._observation_space = gym.spaces.Box(
                low=-np.inf, high=np.inf, shape=(obs_dim,), dtype=np.float32
            )

        if isinstance(self._action_space, gym.spaces.Dict):
            first_action_space = next(iter(self._action_space.spaces.values()))

            if isinstance(first_action_space, gym.spaces.Discrete):
                self._action_space = first_action_space

    def _flatten_dict_obs(self, obs: dict) -> np.ndarray:
        """Flatten a dictionary observation to a 1D numpy array."""
        if not isinstance(obs, dict):
            return obs

        flattened = []
        for key in sorted(obs.keys()):
            value = obs[key]
            if isinstance(value, np.ndarray):
                flattened.append(value.flatten())
            else:
                flattened.append(np.array([value]).flatten())

        return np.concatenate(flattened, dtype=np.float32)

    def _extract_dict_action(self, action) -> dict:
        """Convert flattened action back to dict format if needed."""
        if isinstance(action, dict):
            return action

        if isinstance(self._action_space, gym.spaces.Discrete):
            first_key = next(iter(self._konic_action_space.to_gym().spaces.keys()))
            return {first_key: action}

        return action

    @property
    def action_space(self) -> gym.spaces.Space:
        return self._action_space

    @property
    def observation_space(self) -> gym.spaces.Space:
        return self._observation_space

    @property
    def reward_composer(self) -> KonicRewardComposer:
        return self._reward_composer

    @property
    def termination_composer(self) -> KonicTerminationComposer:
        return self._termination_composer

    def step(self, action: dict[str, Any] | Any) -> KonicStepStateObservation:
        """
        Execute one step in the environment.

        If flatten_spaces=True, this method handles flattening/unflattening automatically.
        Subclasses should implement their step logic in terms of the original Dict spaces.
        """
        if self._flatten_spaces:
            action = self._extract_dict_action(action)

        obs, reward, terminated, truncated, info = self._step(action)

        if self._flatten_spaces and isinstance(obs, dict):
            obs = self._flatten_dict_obs(obs)

        return obs, reward, terminated, truncated, info

    def _step(self, action: dict[str, Any]) -> KonicStepStateObservation:
        """
        Internal step implementation that subclasses should override.
        This method always receives actions in Dict format.
        """
        raise NotImplementedError(
            "step() or _step() is not implemented for this environment. "
            "Override KonicEnvironment._step(self, action) to advance the environment and return "
            "a KonicStepStateObservation (observation, reward, terminated, truncated, info). "
            f"Called with action={action!r}."
        )

    def reset(
        self, seed: int | None = None, options: dict | None = None
    ) -> KonicResetStateObservation:
        """
        Reset the environment.

        If flatten_spaces=True, this method handles flattening automatically.
        Subclasses should implement their reset logic in terms of the original Dict spaces.
        """
        obs, info = self._reset(seed, options)

        if self._flatten_spaces and isinstance(obs, dict):
            obs = self._flatten_dict_obs(obs)

        return obs, info

    def _reset(
        self, seed: int | None = None, options: dict | None = None
    ) -> KonicResetStateObservation:
        """
        Internal reset implementation that subclasses should override.
        This method should return observations in Dict format.
        """
        raise NotImplementedError(
            "reset() or _reset() is not implemented for this environment. "
            "Override KonicEnvironment._reset(self, seed: int | None = None) to initialize "
            "the environment and return a KonicResetStateObservation (observation, info). "
            f"Called with seed={seed!r}."
        )

    def get_obs(self):
        raise NotImplementedError(
            "get_obs() is not implemented for this environment. "
            "Override KonicEnvironment.get_obs() to return the current observation "
            "matching the configured observation_space."
        )

    def get_info(self):
        return {}



"""Agent registration for Konic Cloud Platform.

This module provides the agent registration mechanism used by the
Konic runner to discover and load agents for training.
"""

from __future__ import annotations

import inspect
from typing import TYPE_CHECKING, Any

from konic.common.errors.cli import KonicValidationError

if TYPE_CHECKING:
    from konic.agent import KonicAgent

_REGISTERED_AGENT_INSTANCE: KonicAgent | None = None
_REGISTERED_ENV_CLASS: Any = None


def _perform_registration(agent_instance: KonicAgent, name: str) -> KonicAgent:
    """
    Internal function to perform agent registration.

    Args:
        agent_instance: The KonicAgent instance to register
        name: The name to register the agent under

    Returns:
        The registered agent instance

    Raises:
        KonicValidationError: If validation fails
    """
    global _REGISTERED_AGENT_INSTANCE, _REGISTERED_ENV_CLASS

    if not name:
        raise KonicValidationError("Agent name is required.")

    if inspect.isclass(agent_instance):
        raise KonicValidationError(
            f"Expected an agent instance, got class {agent_instance.__name__}. "
            "Please instantiate it first."
        )

    env_class = agent_instance.get_environment()

    try:
        setattr(
            agent_instance,
            "_konic_meta",
            {"name": name, "env_module": env_class},
        )
    except AttributeError as e:
        raise KonicValidationError(f"Failed to set metadata on agent instance: {e}") from e

    _REGISTERED_AGENT_INSTANCE = agent_instance
    _REGISTERED_ENV_CLASS = env_class

    return agent_instance


def register_agent(agent_instance: Any, name: str) -> Any:
    """
    Register a KonicAgent instance for use with Konic Cloud Platform.

    This function must be called in your agent's entrypoint file to make
    the agent discoverable by the Konic runner.

    Args:
        agent_instance: The KonicAgent instance to register (must be instantiated)
        name: The unique name for this agent

    Returns:
        The registered agent instance

    Raises:
        KonicValidationError: If the agent is a class instead of an instance,
            or if the name is empty

    Example:
        >>> from konic.agent import KonicAgent
        >>> from konic.runtime import register_agent
        >>>
        >>> agent = KonicAgent(environment=MyEnvironment())
        >>> register_agent(agent, name="my-trading-agent")
    """
    return _perform_registration(agent_instance, name)


def get_registered_agent() -> tuple[Any, Any]:
    """
    Get the registered agent and environment class.

    This function is used internally by the Konic runner to retrieve
    the registered agent after loading the agent module.

    Returns:
        A tuple of (agent_instance, environment_class). Both may be None
        if no agent has been registered.
    """
    return _REGISTERED_AGENT_INSTANCE, _REGISTERED_ENV_CLASS



def runner(env):
    print("=" * 60)
    print("Texas Hold'em Poker Environment Demo - 10 Hands")
    print("=" * 60)

    stats = {
        "hands_played": 0,
        "hands_won": 0,
        "hands_lost": 0,
        "hands_tied": 0,
        "total_reward": 0.0,
        "total_pot_won": 0,
        "total_pot_lost": 0,
        "folds": 0,
        "showdowns": 0,
    }

    for hand_num in range(1, 11):
        obs, info = env.reset()

        print(f"\n{'=' * 60}")
        print(f"Hand {hand_num}/10")
        print(f"{'=' * 60}")
        print(f"Player hand: {info['player_hand']}")
        print(f"Starting pot: ${info['pot']}")
        print(f"Player chips: ${info['player_chips']}")
        print()

        hand_reward = 0
        step_count = 0

        while step_count < 20:
            action = env.action_space.sample()
            obs, reward, terminated, truncated, info = env.step(action)
            hand_reward += float(reward)
            step_count += 1

            action_names = ["Fold", "Call/Check", "Raise"]
            print(
                f"  Step {step_count}: {action_names[action['action']]} | "
                f"Round: {info['current_round']} | Pot: ${info['pot']} | "
                f"Reward: {reward:.2f}"
            )

            if terminated or truncated:
                print(f"\n  Hand Result: {info['winner'].upper()}")
                print(f"  Final pot: ${info['pot']}")
                print(f"  Player chips: ${info['player_chips']}")
                print(f"  Hand reward: {hand_reward:.2f}")

                stats["hands_played"] += 1
                stats["total_reward"] += hand_reward

                if info["winner"] == "player":
                    stats["hands_won"] += 1
                    stats["total_pot_won"] += info["pot"]
                elif info["winner"] == "opponent":
                    stats["hands_lost"] += 1
                    if action["action"] == 0:
                        stats["folds"] += 1
                elif info["winner"] == "tie":
                    stats["hands_tied"] += 1

                if info["current_round"] == "SHOWDOWN":
                    stats["showdowns"] += 1

                break

        print(f"\n{'=' * 60}")
        print("FINAL STATISTICS")
        print(f"{'=' * 60}")
        print(f"Hands Played:     {stats['hands_played']}")
        print(
            f"Hands Won:        {stats['hands_won']} ({stats['hands_won']/stats['hands_played']*100:.1f}%)"
        )
        print(
            f"Hands Lost:       {stats['hands_lost']} ({stats['hands_lost']/stats['hands_played']*100:.1f}%)"
        )
        print(
            f"Hands Tied:       {stats['hands_tied']} ({stats['hands_tied']/stats['hands_played']*100:.1f}%)"
        )
        print(f"Folds:            {stats['folds']}")
        print(f"Showdowns:        {stats['showdowns']}")
        print(f"Total Reward:     {stats['total_reward']:.2f}")
        print(f"Average Reward:   {stats['total_reward']/stats['hands_played']:.2f}")
        print(f"Total Pot Won:    ${stats['total_pot_won']}")
        print(f"Final Chips:      ${env.player_chips}")
        print(f"Net Profit/Loss:  ${env.player_chips - env.initial_chips}")
        print("=" * 60)



from typing import Any

import torch
import torch.nn as nn
from ray.rllib.core import Columns
from ray.rllib.core.rl_module.apis import ValueFunctionAPI
from ray.rllib.core.rl_module.torch import TorchRLModule
from ray.rllib.utils.annotations import override

from konic.module.base import BaseTorchModule
from konic.module.type import KonicAlgorithmType


class KonicTorchPPO(BaseTorchModule, TorchRLModule, ValueFunctionAPI):
    algorithm = KonicAlgorithmType.PPO

    @override(TorchRLModule)
    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

    @override(TorchRLModule)
    def setup(self):
        super().setup()

        obs_space = self.config.observation_space
        action_space = self.config.action_space
        hidden_dim = int(self.model_config.get("hidden_dim", 64))

        in_dim = obs_space.shape[0]
        out_dim = int(action_space.n)

        self.pi = nn.Sequential(
            nn.Linear(in_dim, hidden_dim),
            nn.ReLU(),
            nn.Linear(hidden_dim, hidden_dim),
            nn.GELU(),
            nn.Linear(hidden_dim, hidden_dim),
            nn.ReLU(),
            nn.Linear(hidden_dim, out_dim),
        )

        self.vf = nn.Sequential(
            nn.Linear(in_dim, hidden_dim),
            nn.ReLU(),
            nn.Linear(hidden_dim, hidden_dim),
            nn.ReLU(),
            nn.Linear(hidden_dim, 1),
        )

    @override(TorchRLModule)
    def _forward_inference(self, batch, **kwargs):
        with torch.no_grad():
            return self._forward_common(batch)

    @override(TorchRLModule)
    def _forward_exploration(self, batch, **kwargs):
        return self._forward_common(batch)

    @override(TorchRLModule)
    def _forward_train(self, batch, **kwargs):
        return self._forward_common(batch)

    def _forward_common(self, batch):
        """Common forward logic for all forward passes"""
        obs = batch[Columns.OBS]

        if len(obs.shape) == 1:
            obs = obs.unsqueeze(0)

        return {Columns.ACTION_DIST_INPUTS: self.pi(obs)}

    @override(ValueFunctionAPI)
    def compute_values(self, batch: dict[str, Any], embeddings: Any | None = None):
        obs = batch[Columns.OBS]

        if len(obs.shape) == 1:
            obs = obs.unsqueeze(0)

        return self.vf(obs).squeeze(-1)



import pytest

from konic.common.errors import KonicError
from konic.environment.reward import KonicRewardComposer, custom_reward
from konic.environment.reward.reducers import SumReducerStrategy


class MockKonicRewardComposer(KonicRewardComposer):
    """Mock KonicRewardComposer for testing."""

    def reward(self) -> float:
        return 0.0


class TestKonicRewardComposer:
    """Test cases for KonicRewardComposer class."""

    @pytest.fixture
    def mock_reward_composer(self):
        return MockKonicRewardComposer()

    def test_initialization(self, mock_reward_composer):
        """Test KonicRewardComposer initialization."""
        assert mock_reward_composer.reducer == SumReducerStrategy

    def test_set_env(self, mock_reward_composer):
        """Test set_env method."""
        mock_env = object()
        mock_reward_composer.set_env(mock_env)
        assert mock_reward_composer.env is mock_env

    def test_reward_base_method(self, mock_reward_composer):
        """Test base reward method."""
        assert mock_reward_composer.reward() == 0.0

    def test_compose_without_custom_functions(self, mock_reward_composer):
        """Test compose method without custom reward functions."""
        result = mock_reward_composer.compose()
        assert result == 0.0

    def test_compose_with_single_custom_function(self, mock_reward_composer):
        """Test compose with a single custom function returning a value."""

        @custom_reward
        def custom_reward_fn(self) -> float:
            return 10.0

        mock_reward_composer.custom_reward = custom_reward_fn.__get__(
            mock_reward_composer, type(mock_reward_composer)
        )

        result = mock_reward_composer.compose()
        assert result == 10.0

    def test_compose_with_single_custom_function_negative(self, mock_reward_composer):
        """Test compose with a single custom function returning negative value."""

        @custom_reward
        def custom_reward_fn(self) -> float:
            return -5.0

        mock_reward_composer.custom_reward = custom_reward_fn.__get__(
            mock_reward_composer, type(mock_reward_composer)
        )

        result = mock_reward_composer.compose()
        assert result == -5.0

    def test_compose_with_multiple_custom_functions(self, mock_reward_composer):
        """Test compose with multiple custom functions that get summed."""

        @custom_reward
        def custom_reward_1(self) -> float:
            return 10.0

        @custom_reward
        def custom_reward_2(self) -> float:
            return 20.0

        @custom_reward
        def custom_reward_3(self) -> float:
            return 5.5

        mock_reward_composer.custom_reward_1 = custom_reward_1.__get__(
            mock_reward_composer, type(mock_reward_composer)
        )
        mock_reward_composer.custom_reward_2 = custom_reward_2.__get__(
            mock_reward_composer, type(mock_reward_composer)
        )
        mock_reward_composer.custom_reward_3 = custom_reward_3.__get__(
            mock_reward_composer, type(mock_reward_composer)
        )

        result = mock_reward_composer.compose()
        assert result == 35.5

    def test_compose_with_multiple_custom_functions_mixed_signs(self, mock_reward_composer):
        """Test compose with multiple custom functions with mixed positive and negative values."""

        @custom_reward
        def custom_reward_1(self) -> float:
            return 100.0

        @custom_reward
        def custom_reward_2(self) -> float:
            return -30.0

        @custom_reward
        def custom_reward_3(self) -> float:
            return -20.0

        mock_reward_composer.custom_reward_1 = custom_reward_1.__get__(
            mock_reward_composer, type(mock_reward_composer)
        )
        mock_reward_composer.custom_reward_2 = custom_reward_2.__get__(
            mock_reward_composer, type(mock_reward_composer)
        )
        mock_reward_composer.custom_reward_3 = custom_reward_3.__get__(
            mock_reward_composer, type(mock_reward_composer)
        )

        result = mock_reward_composer.compose()
        assert result == 50.0

    def test_compose_with_error_in_custom_function(self, mock_reward_composer):
        """Test compose handles errors in custom reward functions."""

        @custom_reward
        def custom_reward_with_error(self) -> float:
            raise ValueError("Test error")

        mock_reward_composer.custom_reward_with_error = custom_reward_with_error.__get__(
            mock_reward_composer, type(mock_reward_composer)
        )

        with pytest.raises(KonicError, match="Error in custom reward function"):
            mock_reward_composer.compose()

    def test_unimplemented_reward_raises_error(self):
        """Test that calling reward on base composer without implementation raises error."""
        composer = KonicRewardComposer()

        with pytest.raises(KonicError, match="KonicRewardComposer.reward\\(\\) is not implemented"):
            composer.reward()

    def test_custom_decorator_adds_attribute(self):
        """Test that @custom_reward decorator adds the correct attribute."""

        @custom_reward
        def test_function():
            return 1.0

        assert hasattr(test_function, "_konic_custom_reward")
        assert getattr(test_function, "_konic_custom_reward") is True

    def test_compose_accesses_env(self, mock_reward_composer):
        """Test that custom functions can access the environment."""
        mock_env = type("MockEnv", (), {"test_value": 42})()
        mock_reward_composer.set_env(mock_env)

        @custom_reward
        def custom_reward_with_env(self) -> float:
            return float(self.env.test_value)

        mock_reward_composer.custom_reward_with_env = custom_reward_with_env.__get__(
            mock_reward_composer, type(mock_reward_composer)
        )

        result = mock_reward_composer.compose()
        assert result == 42.0

    def test_compose_with_zero_value_custom_functions(self, mock_reward_composer):
        """Test compose with custom functions that return zero."""

        @custom_reward
        def custom_reward_zero(self) -> float:
            return 0.0

        mock_reward_composer.custom_reward_zero = custom_reward_zero.__get__(
            mock_reward_composer, type(mock_reward_composer)
        )

        result = mock_reward_composer.compose()
        assert result == 0.0

    def test_compose_with_float_precision(self, mock_reward_composer):
        """Test compose handles floating point precision correctly."""

        @custom_reward
        def custom_reward_1(self) -> float:
            return 0.1

        @custom_reward
        def custom_reward_2(self) -> float:
            return 0.2

        mock_reward_composer.custom_reward_1 = custom_reward_1.__get__(
            mock_reward_composer, type(mock_reward_composer)
        )
        mock_reward_composer.custom_reward_2 = custom_reward_2.__get__(
            mock_reward_composer, type(mock_reward_composer)
        )

        result = mock_reward_composer.compose()
        assert pytest.approx(result, rel=1e-9) == 0.3

    def test_compose_with_env_dependent_rewards(self, mock_reward_composer):
        """Test compose with multiple custom functions that depend on environment state."""
        mock_env = type("MockEnv", (), {"score": 100, "bonus": 50, "penalty": -10})()
        mock_reward_composer.set_env(mock_env)

        @custom_reward
        def score_reward(self) -> float:
            return float(self.env.score)

        @custom_reward
        def bonus_reward(self) -> float:
            return float(self.env.bonus)

        @custom_reward
        def penalty_reward(self) -> float:
            return float(self.env.penalty)

        mock_reward_composer.score_reward = score_reward.__get__(
            mock_reward_composer, type(mock_reward_composer)
        )
        mock_reward_composer.bonus_reward = bonus_reward.__get__(
            mock_reward_composer, type(mock_reward_composer)
        )
        mock_reward_composer.penalty_reward = penalty_reward.__get__(
            mock_reward_composer, type(mock_reward_composer)
        )

        result = mock_reward_composer.compose()
        assert result == 140.0

    def test_custom_decorator_preserves_function_metadata(self):
        """Test that @custom_reward decorator preserves function name and docstring."""

        @custom_reward
        def my_custom_reward() -> float:
            """This is a test docstring."""
            return 1.0

        assert my_custom_reward.__name__ == "my_custom_reward"
        assert my_custom_reward.__doc__ == "This is a test docstring."

    def test_compose_without_set_env(self, mock_reward_composer):
        """Test compose works when env is not set (for custom functions not using env)."""

        @custom_reward
        def simple_reward(self) -> float:
            return 25.0

        mock_reward_composer.simple_reward = simple_reward.__get__(
            mock_reward_composer, type(mock_reward_composer)
        )

        result = mock_reward_composer.compose()
        assert result == 25.0



import gymnasium as gym

from konic.agent import KonicAgent
from konic.runtime import register_agent

env = gym.make("CartPole-v1")
agent = KonicAgent(environment=env)

register_agent(agent, name="{{cookiecutter.name.replace(' ', '-').lower()}}")



from konic.environment.termination.reducers.base import BaseKonicTerminationReducer
from konic.environment.termination.reducers.strategy import OrReducerStrategy

__all__ = ["BaseKonicTerminationReducer", "OrReducerStrategy"]



from enum import Enum


class BoilerplateOptions(str, Enum):
    basic = "basic"


class TrainingStatus(str, Enum):
    """Training job status enum."""

    PENDING = "pending"
    INITIALIZING = "initializing"
    RUNNING = "running"
    COMPLETED = "completed"
    FAILED = "failed"
    CANCELLED = "cancelled"


class SSEEventType(str, Enum):
    """Server-Sent Events event types for training streams."""

    STATUS = "status"
    METRICS = "metrics"
    HEARTBEAT = "heartbeat"
    ERROR = "error"



from __future__ import annotations

from typing import TYPE_CHECKING

from konic.common.errors import KonicError
from konic.environment.reward.base import BaseKonicRewardComposer
from konic.environment.reward.reducers.base import BaseKonicRewardReducer
from konic.environment.reward.reducers.strategy import SumReducerStrategy
from konic.environment.reward.utils import get_custom_reward_fns

if TYPE_CHECKING:
    from konic.environment import KonicEnvironment


class KonicRewardComposer[TKonicEnvironment: "KonicEnvironment"](BaseKonicRewardComposer):
    """Concrete implementation of the reward composer that uses a reducer strategy."""

    env: TKonicEnvironment
    reducer: type[BaseKonicRewardReducer] = SumReducerStrategy

    def set_env(self, env: TKonicEnvironment):
        """
        Bind the environment instance to this reward composer.

        This method is typically called by KonicEnvironment during initialization
        to provide the reward composer with a reference to the environment instance
        for use in custom reward functions.

        Args:
            env: The environment instance to bind.
        """
        self.env = env

    def compose(self) -> float:
        """Compose the total reward using custom functions and reducer."""
        reducer = self.reducer()
        fns = get_custom_reward_fns(self)

        if any(fns):
            try:
                return reducer.reduce(fns)
            except Exception as e:
                raise KonicError(f"Error in custom reward function: {e}") from e

        return self.reward()

    def reward(self) -> float:
        """Compute the base reward value."""
        raise KonicError(f"{self.__class__.__name__}.reward() is not implemented")
